#include "php.h"
#include "zend_compile.h"
#include "php_ini.h"
#include "php_globals.h"
#include "php_orbit.h"
#include "phorb.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

extern void link_shutdown(void);

static zend_class_entry	*corba_class;
static int le_obj = 0;

void phorb_ensure_orb(CORBA_Environment *ev)
{
	int argc = 1;
	char *argv[] = { "php_orbit", NULL };
	OrbSLS_FETCH();

	if (!OrbSG(orb)) {
		OrbSG(orb) = CORBA_ORB_init(&argc, argv, 
		                            "orbit-local-orb", ev);
		OrbSG(active) = 1;
		OrbSG(counter) = 0;
	}
}

void phorb_shutdown_orb(void)
{
	CORBA_Environment ev;
	OrbSLS_FETCH();

	if (OrbSG(orb)) {
		CORBA_exception_init(&ev);
		CORBA_ORB_destroy(OrbSG(orb), &ev);
		CORBA_Object_release((CORBA_Object)OrbSG(orb), &ev);
		CORBA_exception_free(&ev);
		OrbSG(orb) = NULL;
		
		/*
		 * CORBA_ORB_destroy doesn't shutdown linc2
		 * and we leak file descriptors
		 */
		link_shutdown();
		ORBit_genuid_fini();
	}
}

static void phorb_free_object(void *val)
{
	CORBA_Object_release((CORBA_Object)val, NULL);
	if (!--OrbSG(counter) && !OrbSG(active))
		phorb_shutdown_orb();
}

gboolean phorb_is_object(zval *val)
{
	return Z_TYPE_P(val) == IS_OBJECT && Z_OBJCE_P(val) == corba_class;
}

zval *phorb_init_object(zval *res)
{
	object_init_ex(res, corba_class);

	return res;
}

ulong phorb_register_object(CORBA_Object obj)
{
	OrbSG(counter)++;

	return zend_list_insert((void *)obj, le_obj);
}

void phorb_init_corba_class(zend_class_entry *objclass)
{
	corba_class = objclass;
	le_obj = phorb_register_destructor(phorb_free_object, NULL);
}

void phorb_fini_corba_class(void)
{
}

static gboolean phorb_invoke(CORBA_Object cobj, ORBit_IMethod *method,
                             zval **z_argv, zval *res)
{
	CORBA_Environment ev;
	CORBA_TypeCode ret_tc;
	gpointer ret = NULL, retptr, *argv, *argv_ptr;
	gboolean retcode = FALSE;
	int i;

	phorb_reset_exception_counter();

	ret_tc = method->ret;
	while(ret_tc && ret_tc->kind == CORBA_tk_alias)
		ret_tc = ret_tc->subtypes[0];

	if (ret_tc != CORBA_OBJECT_NIL && ret_tc->kind != CORBA_tk_void) {
		switch(ret_tc->kind) {
		case CORBA_tk_any:
		case CORBA_tk_sequence:
		case CORBA_tk_array:
			retptr = NULL;
			ret = &retptr;
			break;
		case CORBA_tk_struct:
		case CORBA_tk_union:
			if (!(method->flags & ORBit_I_COMMON_FIXED_SIZE)) {
				retptr = NULL;
				ret = &retptr;
				break;
			}
		default:
			ret = ORBit_small_alloc(method->ret);
		}
	}

	argv = g_new0(gpointer, method->arguments._length);
	argv_ptr = g_new0(gpointer, method->arguments._length);

	for(i = 0; i < method->arguments._length; i++) {
		CORBA_TypeCode tc = method->arguments._buffer[i].tc;
		gint flags = method->arguments._buffer[i].flags;

		if (flags & (ORBit_I_ARG_IN | ORBit_I_ARG_INOUT))
			argv[i] = ORBit_small_alloc(tc);
		else {
			argv[i] = &argv_ptr[i];
			if ((tc->kind == CORBA_tk_any
			     || tc->kind == CORBA_tk_sequence
			     || tc->kind == CORBA_tk_struct
			     || tc->kind == CORBA_tk_union
			     || tc->kind == CORBA_tk_array)
			    && (!(flags & ORBit_I_COMMON_FIXED_SIZE)))
				argv_ptr[i] = NULL;
			else
				argv_ptr[i] = ORBit_small_alloc(tc);
		}
	}

	for(i = 0; i < method->arguments._length; i++) {
		gconstpointer buf = argv[i];
		if (!(method->arguments._buffer[i].flags
		      & (ORBit_I_ARG_IN | ORBit_I_ARG_INOUT)))
			continue;

		if (!phorb_put_zval(&buf,
		                    method->arguments._buffer[i].tc,
		                    z_argv[i])) {
			zend_error(E_ERROR,
			           "Could not marshal parameter `%s' (%d)",
			           method->arguments._buffer[i].name
			           	? method->arguments._buffer[i].name
			           	: "<unknown>", i);
			goto cleanup;
		}
	}

	CORBA_exception_init(&ev);
	ORBit_small_invoke_stub(cobj, method, ret, argv,
	                        CORBA_OBJECT_NIL, &ev);
	if (ev._major == CORBA_SYSTEM_EXCEPTION) {
		CORBA_SystemException *ex = ev._any._value;
		zval *tmp;

		tmp = phorb_system_except(ev._id, ex->minor, ex->completed);

		phorb_throw(tmp);
		goto cleanup;
	} else if (ev._major == CORBA_USER_EXCEPTION) {
		CORBA_TypeCode tc = ev._any._type;
		gconstpointer data = ev._any._value;
		zval *tmp = NULL;

		if (tc)
			tmp = phorb_get_zval(&data, tc);
		if (tmp) {
			convert_to_array(tmp);
			tmp = phorb_user_except(ev._id, tmp);
		} else {
			zend_error(E_WARNING,
			           "Unknown exception of type `%s' received",
			           ev._id);
			tmp = phorb_system_except(
			      	"IDL:omg.org/CORBA/UNKNOWN:1.0", 0,
			      	CORBA_COMPLETED_MAYBE);
		}

		phorb_throw(tmp);
		goto cleanup;
	}
	CORBA_exception_free(&ev);

	if (ret_tc != CORBA_OBJECT_NIL && ret_tc->kind != CORBA_tk_void) {
		gconstpointer retptr;
		zval *tmp;

		switch(ret_tc->kind) {
		case CORBA_tk_any:
		case CORBA_tk_sequence:
		case CORBA_tk_array:
			retptr = *(gpointer *)ret;
			break;
		case CORBA_tk_struct:
		case CORBA_tk_union:
			if (!(method->flags & ORBit_I_COMMON_FIXED_SIZE)) {
				retptr = *(gpointer *)ret;
				break;
			}
		default:
			retptr = ret;
		}

		tmp = phorb_get_zval(&retptr, ret_tc);
		if (!tmp) {
			zend_error(E_ERROR,
			           "Could not demarshal return value");
			goto cleanup;
		}

		zval_dtor(res);
		*res = *tmp;
		FREE_ZVAL(tmp);
	}

	for(i = 0; i < method->arguments._length; i++) {
		gconstpointer retptr;
		gint flags = method->arguments._buffer[i].flags;
		zval *tmp;

		if (!(flags & (ORBit_I_ARG_OUT | ORBit_I_ARG_INOUT)))
			continue;

		if (flags & ORBit_I_ARG_OUT)
			retptr = argv_ptr[i];
		else
			retptr = argv[i];

		tmp = phorb_get_zval(&retptr,
		                     method->arguments._buffer[i].tc);
		if (!tmp) {
			zend_error(E_ERROR,
			           "Could not demarshal parameter `%s' (%d)",
			           method->arguments._buffer[i].name
			           	? method->arguments._buffer[i].name
			           	: "<unknown>", i);
			goto cleanup;
		}

		REPLACE_ZVAL_VALUE(&z_argv[i], tmp, 0);
		FREE_ZVAL(tmp);
	}

	retcode = TRUE;

cleanup:
	if (ret) {
		switch(ret_tc->kind) {
		case CORBA_tk_any:
		case CORBA_tk_sequence:
		case CORBA_tk_array:
			CORBA_free(retptr);
			break;
		case CORBA_tk_struct:
		case CORBA_tk_union:
			if (!(method->flags & ORBit_I_COMMON_FIXED_SIZE)) {
				CORBA_free(retptr);
				break;
			}
		default:
			CORBA_free(ret);
		}
	}

	if (argv) {
		for(i = 0; i < method->arguments._length; i++) {
			if (method->arguments._buffer[i].flags
			    & ORBit_I_ARG_OUT)
				CORBA_free(argv_ptr[i]);
			else
				CORBA_free(argv[i]);
		}

		g_free(argv);
		g_free(argv_ptr);
	}

	if (phorb_reset_exception_counter()) {
		phorb_catch();
		return FALSE;
	}

	return retcode;
}

CORBA_char *phorb_get_object_repoid(CORBA_Object cobj, CORBA_Environment *ev)
{
	struct CORBA_Object_internal {
		struct ORBit_RootObject_struct	root;
		void				*reserved;
		GQuark				type_qid;
	};
	CORBA_Environment ev_;
	const char *repoid = NULL;

	if (!ev) {
		ev = &ev_;
		CORBA_exception_init(ev);
	}

	repoid = g_quark_to_string(
	         	((struct CORBA_Object_internal *)cobj)->type_qid);
	/*
	 * FIXME: this doesn't work with non-ORBit2 communication partners
	 *
	repoid = ORBit_small_get_type_id(cobj, ev);
	*/

	if (ev == &ev_)
		CORBA_exception_free(ev);

	if (!repoid)
		return NULL;

	return CORBA_string_dup(repoid);
}

void phorb_invoke_method(INTERNAL_FUNCTION_PARAMETERS,
                         zend_property_reference *props)
{
	CORBA_Environment ev;
	int argc = ZEND_NUM_ARGS();
	zval **argv = NULL, **handle, *pobj = props->object;
	phorbIfaceInfo *info = NULL;
	CORBA_Object cobj;
	ORBit_IMethod *imethod;
	zend_overloaded_element	*fname =
		(zend_overloaded_element *)props->elements_list->tail->data;
	CORBA_char *repo_id;
	char *method;
	int i, type;
	TSRMLS_FETCH();
	OrbSLS_FETCH();

	CORBA_exception_init(&ev);

	convert_to_string(&fname->element);
	method = Z_STRVAL(fname->element);
	if (!method) {
		zend_error(E_WARNING, "Can't make method name a string");
		goto error;
	}

	if (strcmp(method, "corba") == 0
	    && zend_llist_count(props->elements_list) == 1
	    && zend_hash_num_elements(Z_OBJPROP_P(pobj)) == 0) {
		char *name;
		int name_len;
		zval *tmp;

		if (zend_parse_parameters_ex(ZEND_PARSE_PARAMS_QUIET,
		                             argc TSRMLS_CC, "s",
		                             &name, &name_len) == FAILURE) {
			zend_error(E_WARNING, "CORBA constructor expects "
		                              "exactly one string parameter");
			goto ctor_error;
		}

		phorb_ensure_orb(&ev);

		cobj = CORBA_ORB_string_to_object(OrbSG(orb), name, &ev);
		if (!cobj) {
			zend_error(E_WARNING,
			           "Could not map IOR to Object in CORBA()");
			goto ctor_error;
		}

		g_assert(Z_TYPE_P(pobj) == IS_OBJECT);

		MAKE_STD_ZVAL(tmp);
		ZVAL_RESOURCE(tmp, phorb_register_object(cobj));

		zend_hash_index_update(Z_OBJPROP_P(pobj), 0, &tmp,
		                       sizeof(zval *), NULL);

		goto leave;
	} else if (strcmp(method, "_equal") == 0) {
		CORBA_Object cobj2;
		zval *arg;

		if (zend_parse_parameters_ex(ZEND_PARSE_PARAMS_QUIET,
		                             argc TSRMLS_CC, "O",
		                             &arg, corba_class) == FAILURE) {
			zend_error(E_WARNING, "_equal() expects exactly one "
			                      "CORBA parameter");
			goto leave;
		}

		zend_hash_index_find(Z_OBJPROP_P(pobj), 0, (void *)&handle);
		cobj = (CORBA_Object)zend_list_find(Z_RESVAL_PP(handle),
		                                    &type);

		zend_hash_index_find(Z_OBJPROP_P(arg), 0, (void *)&handle);
		cobj2 = (CORBA_Object)zend_list_find(Z_RESVAL_PP(handle),
		                                     &type);

		if (cobj == cobj2) {
			RETVAL_TRUE;
		} else {
			RETVAL_FALSE;
		}

		goto leave;
	}

	argv = (zval **)emalloc(sizeof(zval *) * argc);
	zend_get_parameters_array(ht, argc, argv);

	zend_hash_index_find(Z_OBJPROP_P(pobj), 0, (void *)&handle);
	cobj = (CORBA_Object)zend_list_find(Z_RESVAL_PP(handle), &type);

	repo_id = phorb_get_object_repoid(cobj, &ev);
	if (repo_id) {
		info = phorb_find_interface_description(NULL, repo_id);
		CORBA_free(repo_id);
	}

	if (!info) {
		zend_error(E_WARNING, "Invocation on undefined interface");
		goto error;
	}

	imethod = phorb_find_interface_method(info, method);
	if (!imethod) {
		zend_error(E_WARNING,
		           "Invocation on undefined method `%s' in "
		           "interface `%s'", method, info->iface->tc->repo_id);
		goto error;
	}

	if (argc != imethod->arguments._length) {
		WRONG_PARAM_COUNT;
		goto error;
	}
	for(i = 0; i < argc; i++) {
		if ((imethod->arguments._buffer[i].flags
		     & (ORBit_I_ARG_OUT | ORBit_I_ARG_INOUT))
		    && !PZVAL_IS_REF(argv[i])) {
			zend_error(E_WARNING, "Parameter %d must be passed "
			                      "by reference", i);
			goto error;
		}
	}

	if (!phorb_invoke(cobj, imethod, argv, return_value) < 0)
		goto error;

  leave:
	zval_dtor(&fname->element);
	if (argv)
		efree(argv);
	CORBA_exception_free(&ev);
	return;

  ctor_error:
	zval_dtor(pobj);
	ZVAL_NULL(pobj);
	goto leave;

  error:
  	zval_dtor(return_value);
	RETVAL_NULL();
	goto leave;
}

zval phorb_get_attribute(zend_property_reference *props)
{
	zval **handle;
	zval dres;
	zval *res = &dres;
	zval *pobj = props->object;
	CORBA_Object cobj;
	ORBit_IMethod *imethod;
	zend_overloaded_element	*aname =
		(zend_overloaded_element *)props->elements_list->tail->data;
	CORBA_char *repo_id = NULL;
	char *attribute;
	gchar *method;
	phorbIfaceInfo *info = NULL;
	int type;

	INIT_ZVAL(dres);

	convert_to_string(&aname->element);
	attribute = Z_STRVAL(aname->element);
	if (!attribute) {
		zend_error(E_WARNING, "Can't make attribute name a string");
		goto out;
	}

	zend_hash_index_find(Z_OBJPROP_P(pobj), 0, (void *)&handle);
	cobj = (CORBA_Object)zend_list_find(Z_RESVAL_PP(handle), &type);

	repo_id = phorb_get_object_repoid(cobj, NULL);
	if (repo_id)
		info = phorb_find_interface_description(NULL, repo_id);

	if (!info) {
		zend_error(E_WARNING, "Get invocation on undefined interface");
		goto out;
	}

	if (strcmp(attribute, "_tc") == 0) {
		CORBA_TypeCode tc;
		zval *val;

		tc = phorb_find_typecode(NULL, repo_id);
		if (!tc)
			tc = TC_CORBA_Object;

		res = phorb_init_typecode(res);
		if (!res)
			goto out;

		MAKE_STD_ZVAL(val);
		ZVAL_RESOURCE(val,
		              phorb_register_typecode(
		              	(CORBA_TypeCode)CORBA_Object_duplicate(
		              		(CORBA_Object)tc, NULL)));
		zend_hash_index_update(Z_OBJPROP_P(res), 0, &val,
		                       sizeof(zval *), NULL);

		goto out;
	} else if (strcmp(attribute, "_hash") == 0) {
		CORBA_Environment ev;

		CORBA_exception_init(&ev);
		ZVAL_LONG(res, CORBA_Object_hash(cobj, LONG_MAX, &ev));
		CORBA_exception_free(&ev);

		goto out;
	}

	method = g_strdup_printf("_get_%s", attribute);
	imethod = phorb_find_interface_method(info, method);
	g_free(method);
	if (!imethod || imethod->arguments._length != 0
	    || imethod->ret == CORBA_OBJECT_NIL
	    || imethod->ret->kind == CORBA_tk_void) {
		zend_error(E_WARNING, "Get invocation on undefined attribute "
		           "%s in interface %s", attribute,
		           info->iface->tc->repo_id);
		goto out;
	}

	if (!phorb_invoke(cobj, imethod, NULL, res))
		goto out;

out:
	if (repo_id)
		CORBA_free(repo_id);

	zval_dtor(&aname->element);
	return dres;
}

int phorb_set_attribute(zend_property_reference *props, zval *value)
{
	zval **handle;
	zval *argv[1] = { value };
	zval *pobj = props->object;
	CORBA_Object cobj;
	ORBit_IMethod *imethod;
	zend_overloaded_element	*aname =
		(zend_overloaded_element *)props->elements_list->tail->data;
	CORBA_char *repo_id;
	char *attribute = Z_STRVAL(aname->element);
	gchar *method;
	phorbIfaceInfo *info = NULL;
	int type;

	convert_to_string(&aname->element);
	attribute = Z_STRVAL(aname->element);
	if (!attribute) {
		zend_error(E_WARNING, "Can't make attribute name a string");
		goto error;
	}

	zend_hash_index_find(Z_OBJPROP_P(pobj), 0, (void *)&handle);
	cobj = (CORBA_Object)zend_list_find(Z_RESVAL_PP(handle), &type);

	repo_id = phorb_get_object_repoid(cobj, NULL);
	if (repo_id) {
		info = phorb_find_interface_description(NULL, repo_id);
		CORBA_free(repo_id);
	}

	if (!info) {
		zend_error(E_WARNING, "Set invocation on undefined interface");
		goto error;
	}

	method = g_strdup_printf("_set_%s", attribute);
	imethod = phorb_find_interface_method(info, method);
	g_free(method);
	if (!imethod || imethod->arguments._length != 1
	    || (imethod->ret != CORBA_OBJECT_NIL
	        && imethod->ret->kind != CORBA_tk_void)) {
		zend_error(E_WARNING, "Set invocation on undefined attribute "
		           "%s in interface %s", attribute,
		           info->iface->tc->repo_id);
		goto error;
	}

	SEPARATE_ZVAL(&argv[0]);
	if (!phorb_invoke(cobj, imethod, argv, NULL))
		goto error;

	zval_dtor(&aname->element);
	return SUCCESS;

  error:
	zval_dtor(&aname->element);
	return FAILURE;
}

void phorb_serialize_sleep(INTERNAL_FUNCTION_PARAMETERS)
{
	CORBA_Environment ev;
	CORBA_Object cobj;
	CORBA_char *ior;
	zval **handle;
	int type;
	OrbSLS_FETCH();

	CORBA_exception_init(&ev);
	zend_hash_index_find(Z_OBJPROP_P(this_ptr), 0, (void *)&handle);
	cobj = (CORBA_Object)zend_list_find(Z_RESVAL_PP(handle), &type);

	phorb_ensure_orb(&ev);

	array_init(return_value);
	ior = CORBA_ORB_object_to_string(OrbSG(orb), cobj, &ev);

	add_property_string(this_ptr, "_ior", ior, 1);
	add_next_index_string(return_value, "_ior", 1);

	CORBA_free(ior);
	CORBA_exception_free(&ev);
}

void phorb_serialize_wakeup(INTERNAL_FUNCTION_PARAMETERS)
{
	CORBA_Environment ev;
	CORBA_Object cobj;
	zval *handle, **ior;
	int type;
	OrbSLS_FETCH();

	zend_hash_find(Z_OBJPROP_P(this_ptr), "_ior", 5, (void *)&ior);
	if (!ior) {
		zend_error(E_ERROR, "Bad serialization of CORBA object, "
		                    "no IOR found");
		RETURN_FALSE;
	}
	if (Z_TYPE_PP(ior) != IS_STRING) {
		zend_error(E_ERROR, "Bad serialization of CORBA object, "
		                    "IOR not a string");
		RETURN_FALSE;
	}

	CORBA_exception_init(&ev);
	phorb_ensure_orb(&ev);
	cobj = CORBA_ORB_string_to_object(OrbSG(orb), Z_STRVAL_PP(ior), &ev);
	CORBA_exception_free(&ev);

	if (!cobj) {
		zend_error(E_WARNING, "Could not map IOR to Object while "
		                      "unserializing CORBA Object");
		RETURN_FALSE;
	}
	zval_ptr_dtor(ior);
	zend_hash_del(Z_OBJPROP_P(this_ptr), "_ior", 5);

	MAKE_STD_ZVAL(handle);
	ZVAL_RESOURCE(handle, phorb_register_object(cobj));

	zend_hash_index_update(Z_OBJPROP_P(this_ptr), 0, &handle,
	                       sizeof(zval *), NULL);

	RETURN_TRUE;
}
